#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for carta_envoy.connect
"""
from unittest.mock import patch

import pytest
from carta_envoy.connect import (
    ArchiveConnect,
    NotificationConnect,
    RedisConnect,
    WorkflowConnect,
)
from carta_envoy.utilities import generate_random_str

# pylint: disable=E0401, C0103, C0301, C0415, E1101, E1120, R0201, R0903, W0621


test_settings = {
    "timeout": 1,
    "carta_path": "/fake/path/to/nowhere",
    "redis_host": "noplace.nrao.edu",
    "redis_port": "6397",
    "redis_password": "password",
    "reverse_proxy": "black-abyss.nrao.edu",
    "user_email": "rusty.shackleford@nowhere.nrao.edu",
    "data_location": ".",
    "useCarta": False,
    "notification_url": "http://notification:3458",
    "workflow_url": "http://workflow:3456",
    "wf_request_id": "-1",
}

redis_connect = RedisConnect(settings=test_settings)
notify_connect = NotificationConnect(settings=test_settings)
archive_connect = ArchiveConnect(settings=test_settings)
workflow_connect = WorkflowConnect(settings=test_settings)

CARTA_POST = "carta_envoy.connect.requests.post"


@pytest.fixture
def redis():
    """
    Pytest fixture that provides a sandbox Redis server for testing purposes

    :return Sandbox Redis server
    """
    import fakeredis

    return fakeredis.FakeStrictRedis()


class TestRedisConnect:
    "Test our connection to Redis server"

    @pytest.mark.skip("not yet implemented")
    def test_generate_ids(self):
        """
        TODO

        :return:
        """

    @pytest.mark.skip("not yet implemented")
    def test_generate_url(self):
        """
        TODO

        :return:
        """

    def test_get_available_port(self):
        """
        Tests that get_available_port returns an available port on localhost
        """
        import socket

        with socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM) as s:
            try:
                s.bind(("", redis_connect.get_available_port()))
            except OSError as e:
                # Port not available
                print(e)
                assert False

    @pytest.mark.skip("not yet implemented")
    def test_prepare_redis(self):
        """
        TODO

        :return:
        """

    def test_connect_to_redis_error(self):
        """
        Tests that a connection to an invalid Redis server correctly throws an exception
        """
        import redis

        with pytest.raises(redis.ConnectionError):
            redis_connect.connect_to_redis()

    def test_get_redis_values(self):
        """
        Tests that filled out Redis values are given correctly
        """
        random_id = generate_random_str()
        session_port = 7777
        wrapper_port = 6464
        reverse_proxy_host = "test"
        front_end = f"carta-front-{random_id}"
        back_end = f"carta-back-{random_id}"
        values = [
            f"traefik/http/routers/{front_end}/rule",
            f"traefik/http/routers/{front_end}/service",
            f"traefik/http/services/{front_end}/loadbalancer/servers/0/url",
            f"traefik/http/routers/{front_end}/middlewares/0",
            f"traefik/http/routers/{back_end}/rule",
            f"traefik/http/routers/{back_end}/service",
            f"traefik/http/services/{back_end}/loadbalancer/servers/0/url",
            f"traefik/http/routers/{back_end}/middlewares/0",
            f"traefik/http/routers/{back_end}/middlewares/1",
        ]
        assert (
            value
            in redis_connect.get_redis_values(
                reverse_proxy_host, random_id, random_id, session_port, wrapper_port
            ).keys()
            for value in values
        )

    def test_check_for_duplicate_values(self, redis):
        """
        Tests that duplicate Redis values are correctly reported

        :param redis: Custom fixture that provides a sandbox Redis server
        """
        redis_connect.conn = redis
        redis_connect.generated_ids = {"session_id": "5678efgh", "wrapper_id": "9876jklm"}
        redis.set("duplicate", "value")
        with patch("carta_envoy.connect.RedisConnect.get_redis_values") as values:
            redis_connect.check_for_duplicate_values({"duplicate": "value"}, 9897, 6543)
            assert values.call_count == 1
            values.call_count = 0
            redis_connect.check_for_duplicate_values({"not_duplicate": "value"}, 9897, 6543)
            assert values.call_count == 0


class TestArchiveConnect:
    """Test sending CARTA URL to archive request handler"""

    def test_send_carta_url_to_rh(self):
        with patch(CARTA_POST) as mocked_post:
            archive_connect.send_carta_url_to_rh("http://fake.url/for/testing")
            mocked_post.assert_called_with(
                f"{test_settings['workflow_url']}/workflows/carta/requests/{test_settings['wf_request_id']}/send-url-to-aat",
                json={"carta_url": "http://fake.url/for/testing"},
            )


class TestNotificationConnect:
    """Test sending session-ready and session-expired messages"""

    def test_send_session_ready(self):
        with patch(CARTA_POST) as mocked_post:
            url = "I am carta"
            notify_connect.send_session_ready(wrapper_url=url)
            mocked_post.assert_called_with(
                f"{test_settings['notification_url']}/notify/carta_ready/send",
                json={"destination_email": test_settings["user_email"], "carta_url": url},
            )

    def test_send_session_expired(self):
        with patch(CARTA_POST) as mocked_post:
            notify_connect.send_session_expired()
            mocked_post.assert_called_with(
                f"{test_settings['notification_url']}/notify/carta_expired/send",
                json={
                    "destination_email": test_settings["user_email"],
                },
            )


class TestWorkflowConnect:
    """Test sending CARTA URL to Workspaces systesm"""

    def test_send_carta_url(self):
        with patch(CARTA_POST) as mocked_post:
            url = "my_very_own_special_CARTA_session"
            workflow_connect.send_carta_url(url)
            mocked_post.assert_called_with(
                f"{test_settings['workflow_url']}/workflows/carta/requests/{test_settings['wf_request_id']}/url",
                json={"carta_url": url},
            )
